from flask import current_app
from flask_restful import Resource, Api

from . import api
from ..core.engine import Engine
from ..utils.log import logger

api = Api(api, catch_all_404s=True)


class DagAPI(Resource):
    """
    The api blue
    print to execute etl task
    """

    def __init__(self):
        app_name = current_app.name
        app_context = current_app.config[app_name]
        self.engine = Engine.instance(app_context, detach=True)

    def post(self):
        # data in string format and you have to parse into dictionary
        from flask import request

        try:
            task_args = request.get_json()
            if not task_args:
                task_args = {}
        except Exception as e:
            logger.exception(str(e))
            task_args = {}

        task_args_ = dict(map(lambda x: (x[0], x[1][0]), filter(lambda x: len(x[1]) == 1, dict(request.args).items())))

        if task_args_:
            task_args.update(task_args_)

        tasks = set(task_args.get('tasks', []))
        self._process(*tasks)

    def _process(self, *tasks):
        self.engine.submit_dag(*tasks)
        resp = {"code": 200, "message": "tasks {} scheduled".format(set(tasks))}
        return resp

api.add_resource(DagAPI, '/api/dag')



from . import ParadeCommand
from ..core.engine import Engine
from ..utils.log import logger


class ExecCommand(ParadeCommand):
    """
    The exec command to run a flow or a set tasks,
    if the tasks to execute have dependencies on each other,
    parade will handle them correctly
    """
    requires_workspace = True

    def run_internal(self, context, **kwargs):
        engine = Engine(context)

        task = kwargs.get('task')
        logger.debug('prepare to execute task {}'.format(task))

        engine.execute(task)

    def short_desc(self):
        return 'execute a flow or a set of tasks'

    def config_parser(self, parser):
        parser.add_argument('task', help='the task to execute')



import datetime
import time

import pytz


def datetime_str_to_timestamp(datetime_str, tz=None):
    dt = datetime.datetime.strptime(datetime_str, '%Y-%m-%d %H:%M:%S')
    return datetime_to_timestamp(dt, tz)


def datetime_to_timestamp(dt, tz=None):
    if tz:
        dt = pytz.timezone(tz).localize(dt)
        return int((dt - datetime.datetime(1970, 1, 1, tzinfo=pytz.utc)).total_seconds())
    return int(time.mktime(dt.timetuple()))


def timestamp_to_datetime(ts, tz=None):
    if tz:
        return datetime.datetime.fromtimestamp(ts, pytz.timezone(tz))
    return datetime.datetime.fromtimestamp(ts)



from collections import defaultdict

from ..core.task import Task
from ..utils.modutils import get_class, iter_classes
from ..connection import Connection, Datasource


class Context(object):
    """
    The executor context to support the ETL job executed by the engine
    """
    def __init__(self, name, conf, workdir=None):
        self.name = name
        self.workdir = workdir
        self.conf = conf

        self._conn_cache = defaultdict(Connection)
        self.task_dict = self._get_tasks()

    def setup(self):
        raise NotImplementedError

    def _get_tasks(self):
        """
        generate the task dict [task_key => task_obj]
        :return:
        """
        d = {}
        for task in iter_classes(Task, self.name + '.task'):
            task_name = task.__module__.split('.')[-1]
            d[task_name] = task()
        return d

    def get_connection(self, conn_key):
        """
        Get the connection with the connection key
        :param conn_key: the key of the connection
        :return: the connection instance
        """

        if conn_key not in self._conn_cache:
            conn_conf = self.conf['connection']
            assert conn_key in conn_conf.to_dict(), 'connection {} is not configured'.format(conn_key)
            datasource = Datasource(**conn_conf[conn_key].to_dict())
            conn_cls = get_class(datasource.driver, Connection, 'parade.connection', self.name + '.contrib.connection')
            if conn_cls:
                self._conn_cache[conn_key] = conn_cls(datasource)

        if conn_key not in self._conn_cache:
            raise NotImplementedError("The driver [%s] is not supported".format("conn_key"))

        return self._conn_cache[conn_key]



from ..core.task import Task
from ..utils.log import logger


class Scheduler(object):
    """
    The scheduler to support task-flow scheduling
    """

    @staticmethod
    def parse_forests(*tasks):
        """
        util method to build flow forest
        :param tasks: the specified tasks to form the flow
        :return: the tasks are not dependent by others
        """
        task_roots = dict()
        try:
            for task in tasks:
                assert isinstance(task, Task), 'Invalid task-module [{}] provided'.format(task.__class__.__name__)
                for dep in task.deps:
                    task_roots[dep] = task.name
        except RuntimeError as e:
            logger.exception(str(e))

        forests = list(filter(lambda x: x.name not in task_roots, tasks))
        logger.debug('Task forests generated [{}]'.format(', '.join(map(lambda x: x.name, forests))))
        return forests

    def schedule(self, *tasks, cron=None, flow_name=None):
        """
        schedule the task-flow
        :param flow_name: the flow name
        :param cron: the cron string to schedule the flow
        :return:
        """
        raise NotImplemented

    def unsched(self, flow_name):
        """
        unschedule the task-flow
        :param flow_name: the flow name
        :return:
        """
        pass

    def __init__(self, context):
        self.context = context



import os
import sys
import argparse

from parade.config import ConfigRepo
from parade.command import ParadeCommand
from parade.utils.modutils import iter_classes
from parade.utils.workspace import inside_workspace, load_bootstrap


def _get_commands(in_workspace):
    d = {}
    for cmd in iter_classes(ParadeCommand, 'parade.command', class_filter=lambda cls: cls != ParadeCommand):
        if in_workspace or not cmd.requires_workspace:
            cmd_name = cmd.__module__.split('.')[-1]
            d[cmd_name] = cmd()
    return d


def _get_config_repo(driver, uri, workdir=None):
    for repo in iter_classes(ConfigRepo, 'parade.config', class_filter=lambda cls: cls != ConfigRepo):
        repo_name = repo.__module__.split('.')[-1]
        if repo_name == driver:
            return repo(uri, workdir=workdir)


def execute():
    # load the commands and parse arguments
    parser = argparse.ArgumentParser(description='The CLI of parade engine.')
    inworkspace = inside_workspace()
    sub_parsers = parser.add_subparsers(dest='command')
    cmds = _get_commands(inworkspace)
    for cmdname, cmd in cmds.items():
        cmd_parser = sub_parsers.add_parser(cmdname, help=cmd.help())
        cmd.config_parser(cmd_parser)
    args = parser.parse_args()

    if not args.command:
        parser.print_usage()
        sys.exit(1)

    command = cmds[args.command]
    command_args = args.__dict__

    if inworkspace and command.requires_workspace:
        bootstrap = load_bootstrap()
        workspace_settings = bootstrap['workspace']
        config_settings = bootstrap['config']

        config_driver = config_settings['driver']
        config_uri = config_settings['uri']

        config_repo = _get_config_repo(config_driver, config_uri, workspace_settings['path'])

        config_name = config_settings['name']
        config_profile = config_settings['profile']
        config_version = config_settings['version']

        # add this to enable switch profile via environment
        config_profile = os.environ.get('PARADE_PROFILE', config_profile)

        config = config_repo.load(app_name=config_name, profile=config_profile, version=config_version)

        command_args.update(workspace=workspace_settings['name'], workdir=workspace_settings['path'])
        command_args.update(config=config)

    command.run(**command_args)


if __name__ == '__main__':
    execute()



import os
import sys
import yaml


def inside_workspace():
    return bool(locate_bootstrap())


def locate_bootstrap():
    """Return the path to the closest parade bootstrap file in current directory
    """
    clue = os.environ.get('PARADE_WORKSPACE', '.')
    path = os.path.abspath(clue)
    bootfile = os.path.join(path, 'parade.bootstrap.yml')
    if os.path.exists(bootfile):
        return bootfile
    return None


def load_bootstrap(addpath=True):
    """Initialize environment to use command-line tool from inside a project
    dir. This sets the Scrapy settings module and modifies the Python path to
    be able to locate the project module.
    """
    bootfile = locate_bootstrap()
    if not bootfile:
        return []
    workspace = os.path.dirname(bootfile)
    if addpath and workspace not in sys.path:
        sys.path.append(workspace)
    with open(bootfile) as f:
        content = f.read()
        config_dict = yaml.load(content)
        config_dict['workspace']['path'] = workspace
        return config_dict
    return []



from ..utils.modutils import get_class, iter_classes
from ..scheduler import Scheduler
from . import ParadeCommand


class SchedCommand(ParadeCommand):

    requires_workspace = True

    def run_internal(self, context, **kwargs):
        config = context.conf
        scheduler_driver = config['scheduler.driver']
        scheduler_cls = get_class(scheduler_driver, Scheduler, 'parade.scheduler', context.name + ".contrib.scheduler")
        scheduler = scheduler_cls(context)

        tasks = kwargs.get('task')
        flow_name = kwargs.get('flow_name')
        if not flow_name:
            flow_name = context.name
        cron = kwargs.get('cron')

        scheduler.schedule(*tasks, flow_name=flow_name, cron=cron)

    def short_desc(self):
        return 'schedule a flow with a set of tasks'

    def config_parser(self, parser):
        parser.add_argument('--cron', dest='cron', help='set the schedule settings')
        parser.add_argument('--flow', dest='flow_name', help='set the flow name to schedule')
        parser.add_argument('task', nargs='*', help='the task to schedule')




from . import ParadeCommand


def _create_app(name, context):
    from flask import Flask
    from flask_cors import CORS

    app = Flask(name)
    CORS(app)

    app.config[name] = context

    from ..api import api as api_blueprint
    app.register_blueprint(api_blueprint)

    return app


class ServerCommand(ParadeCommand):
    requires_workspace = True

    def run_internal(self, context, **kwargs):
        port = int(kwargs.get('port', 5000))
        app = _create_app(context.name, context)

        app.run(host="0.0.0.0", port=port, debug=True)

    def short_desc(self):
        return 'start a parade api server'

    def config_parser(self, parser):
        parser.add_argument('-p', '--port', default=5000, help='the port of parade server')



import requests
import yaml

from . import ConfigRepo


class YamlConfig(ConfigRepo):
    def load_internal(self, app_name, profile='default', **kwargs):
        """
        Load the config from a spring cloud server.
        :param app_name: the app name of the configuration
        :param profile: the profile of the configuration, e.g., the deploy environment, like `dev`, `rc`, `prod`, etc.
        :param version: the version of the configuration.
        :return: the loaded configuration object.
        """
        version = kwargs.get('version', 'master')
        uri = self.url.format(name=app_name, profile=profile, version=version)
        if uri.startswith('http://') or uri.startswith('https://'):
            r = requests.get(uri)
            if r.status_code == 200:
                return yaml.load(r.content)
        import os
        if not os.path.isabs(uri):
            workdir = self.kwargs.get('workdir', '.')
            uri = os.path.join(workdir, uri)
        with open(uri, 'r') as f:
            content = f.read()
            return yaml.load(content)

        return dict()



import re
import string
from importlib import import_module
from os.path import join, exists, abspath
from shutil import move, ignore_patterns, copy

import parade
from parade.utils.misc import copytree
from . import ParadeCommand
from ..utils.template import string_camelcase, render_templatefile


class GenTaskCommand(ParadeCommand):
    """
    The init command will create a workspace with template.
    The workspace will hold all user's work (tasks and flows)
    and customized configurations.
    """
    requires_workspace = True

    def short_desc(self):
        return 'generate a task skeleton with specified type'

    def config_parser(self, parser):
        parser.add_argument('task', nargs='*', help='the name of the task to generate')
        parser.add_argument('-t', '--task_type', dest='task_type', help='the type of the task to generate', required=True)

    @staticmethod
    def _is_valid_name(workspace_name):
        """
        check if the workspace name is valid
        :param workspace_name: the specified workspace name
        :return:
        """

        def _module_exists(module_name):
            try:
                import_module(module_name)
                return True
            except ImportError:
                return False

        if not re.search(r'^[_a-zA-Z]\w*$', workspace_name):
            print('Error: Workspace names must begin with a letter and contain' \
                  ' only\nletters, numbers and underscores')
        elif _module_exists(workspace_name):
            print('Error: Module %r already exists' % workspace_name)
        else:
            return True
        return False

    def run_internal(self, context, **kwargs):
        task_names = kwargs['task']
        task_type = kwargs['task_type']

        target_path = join(context.workdir, context.name, "task")

        source_tpl = task_type + ".py.tpl"

        for task_name in task_names:
            target_tpl = task_name + ".py"

            source_tplfile = join(self.templates_dir, source_tpl)
            target_tplfile = join(target_path, target_tpl)
            copy(source_tplfile, target_tplfile)

            render_templatefile(target_tplfile, TaskName=string_camelcase(task_name))

    @property
    def templates_dir(self):
        _templates_base_dir = join(parade.__path__[0], 'template')
        return join(_templates_base_dir, 'task')




import pandas as pd
from sqlalchemy import create_engine, text

from parade.utils.log import logger
from . import Connection, Datasource


class RDBConnection(Connection):
    def __init__(self, datasource):
        Connection.__init__(self, datasource)
        assert isinstance(datasource, Datasource), 'Invalid connection provided'
        # assert connection.host is not None, 'host of connection is required'
        # assert connection.port is not None, 'port of connection is required'
        # assert connection.db is not None, 'db of connection is required'
        # assert connection.protocol is not None, 'protocol of connection is required'

    def open(self):
        uri = self.datasource.uri
        if uri is None:
            authen = None
            uripart = self.datasource.host + ':' + str(self.datasource.port) + '/' + self.datasource.db
            if self.datasource.user is not None:
                authen = self.datasource.user
            if authen is not None and self.datasource.password is not None:
                authen += ':' + self.datasource.password
            if authen is not None:
                uripart = authen + '@' + uripart
            uri = self.datasource.protocol + '://' + uripart
        pandas_conn = create_engine(uri, encoding="utf-8")
        return pandas_conn

    def load(self, table, **kwargs):
        return self.load_query('select * from {}'.format(table))

    def load_query(self, query, **kwargs):
        conn = self.open()
        return pd.read_sql_query(query, con=conn)

    def store(self, df, table, **kwargs):
        assert isinstance(df, pd.DataFrame), "Invalid data type"
        if_exists = kwargs['if_exists'] if 'if_exists' in kwargs else 'fail'
        chunksize = kwargs['chunksize'] if 'chunksize' in kwargs else None
        schema = None
        if table.find('.') >= 0:
            toks = table.split('.', 1)
            schema = toks[0]
            table = toks[1]

        floatcolumns = list(df.select_dtypes(include=['float64', 'float']).keys())
        if len(floatcolumns) > 0:
           logger.warn(
                        "Detect columns with float types {}, you better check if this is caused by NAN-integer column issue of pandas!".format(
                                list(floatcolumns)))

        typehints = dict()
        objcolumns = list(df.select_dtypes(include=['object']).keys())

        if len(objcolumns) > 0:
           logger.warn(
                   "Detect columns with object types {}, which is automatically converted to *VARCHAR(256)*, you can override this by specifying type hints!".format(
                                list(objcolumns)))
        import sqlalchemy.types as sqltypes
        typehints.update(dict((k, sqltypes.VARCHAR(256)) for k in objcolumns))

        # TODO: upddate typehints with user-specified one
        _typehints = kwargs.get('typehints', {})
        from parade.type import stdtype_to_sqltype
        for col, stdtype in _typehints.items():
            logger.info("Column [{}] is set to type [{}]".format(col, str(stdtype)))
            typehints[col] = stdtype_to_sqltype(stdtype)

        def _chunks(_df, _chunksize):
            """Yield successive n-sized chunks from l."""
            for i in range(0, len(_df), _chunksize):
                yield df[i:i + _chunksize]

        for idx, chunk in enumerate(_chunks(df, chunksize)):
            if_exists_ = if_exists
            if idx > 0:
                if_exists_ = 'append'
            chunk.to_sql(name=table, con=self.open(), index=False, schema=schema, if_exists=if_exists_,
                         #chunksize=chunksize,
                         dtype=typehints)
            logger.info("Write rows #{}-#{}".format(idx * chunksize, (idx + 1) * chunksize))

    def last_record(self, task_name):
        _conn = self.open()
        sql = text("""
            SELECT * FROM _task_record
                WHERE task=:task_name AND status=1 ORDER BY create_time desc LIMIT 1
        """)

        _last_record = _conn.execute(sql, task_name=task_name).fetchone()

        if _last_record is not None:
            # return _raw_last_checkpoint[0].strftime("%Y-%m-%d %H:%M:%S")
            return dict(_last_record)
        return None

    def create_record(self, task_name, new_checkpoint):
        _conn = self.open()
        from sqlalchemy import Table
        from sqlalchemy import MetaData
        # TODO append方式下回收超出本次checkpoint的数据
        # if self.get_target_mode() == 'append':
        #     target_table = Table(self.get_target_table(), MetaData(), autoload=True, autoload_with=target_conn)
        #     clear_ins = target_table.delete(whereclause="TIMESTAMP(" + self.checkpoint_column + ") >= '" + self._last_checkpoint + "'")
        #     target_conn.execute(clear_ins)
        # 创建待提交checkpoint
        task_table = Table('_task_record', MetaData(), autoload=True, autoload_with=_conn)
        ins = task_table.insert().values(task=task_name, checkpoint=new_checkpoint)
        return _conn.execute(ins).inserted_primary_key[0]

    def commit_record(self, txn_id):
        _conn = self.open()
        sql = text("""
            UPDATE _task_record SET status = 1, commit_time = now(), update_time = now()
            WHERE id = :id
        """)
        _conn.execute(sql, id=txn_id)

    def rollback_record(self, txn_id, err):
        _conn = self.open()
        sql = text("""
            UPDATE _task_record SET status = 2, message = :err, update_time = now()
            WHERE id = :id
        """)
        _conn.execute(sql, err=str(err), id=txn_id)



import time

from ..connection import Connection
from ..connection.rdb import RDBConnection
from ..utils.log import logger
from ..utils.timeutils import datetime_str_to_timestamp, timestamp_to_datetime


class Task(object):
    """
    The task object executed by the parade engine
    """

    DEFAULT_CHECKPOINT = '1970-01-01 00:00:00'

    def __init__(self):
        """
        _result: the cached task result after execution
        _attributes: the attribute of the task
        :return:
        """
        self._result = None
        self._attributes = {}
        self._last_checkpoint = self.DEFAULT_CHECKPOINT
        self._checkpoint = self.DEFAULT_CHECKPOINT

    @property
    def name(self):
        """
        get the identifier of the task, the default is the class name of task
        :return: the task identifier
        """
        # class_name = self.__class__.__name__
        # task_name = ''
        # for _s_ in class_name:
        #     task_name += _s_ if _s_.islower() else '_' + _s_.lower()
        # return task_name[task_name.startswith("_") and 1:]
        return self.__module__.split('.')[-1]

    @property
    def deps(self):
        return set()

    @property
    def attributes(self):
        """
        the attributes to transferred to the task result
        :return:
        """
        return self._attributes

    def set_attribute(self, key, val):
        self._attributes[key] = val

    @property
    def checkpoint_round(self):
        """
        the time interval the checkpoint will align to
        default value is 1 day
        :return:
        """
        return 3600 * 24

    @property
    def checkpoint_timezone(self):
        """
        the timezone used when recording checkpoint
        default: None, use the local timezone
        :return:
        """
        return None

    @property
    def checkpoint_conn(self):
        """
        the connection to record the checkpoint
        default value is the target connection
        :return:
        """
        return self.target_conn

    # @property
    # def checkpoint_column(self):
    #     """
    #     the column to use as the clue for checkpoint
    #     :return:
    #     """
    #     if self.target_mode == 'append':
    #         raise NotImplementedError
    #     return None

    def _start(self, context):
        """
        start a checkpoint transaction before executing the task
        :param context:
        :return:
        """
        checkpoint_conn = context.get_connection(self.checkpoint_conn)
        assert isinstance(checkpoint_conn, RDBConnection)
        last_record = checkpoint_conn.last_record(self.name)
        if last_record:
            self._last_checkpoint = last_record['checkpoint'].strftime('%Y-%m-%d %H:%M:%S')

        # 基于当前时间,进行粒度对齐后计算本次执行的checkpoint
        now_ts = int(time.time())
        init_ts = datetime_str_to_timestamp(self.DEFAULT_CHECKPOINT, tz=self.checkpoint_timezone)
        checkpoint_ts = now_ts - (now_ts - init_ts) % self.checkpoint_round
        self._checkpoint = timestamp_to_datetime(checkpoint_ts).strftime('%Y-%m-%d %H:%M:%S')

        if self._checkpoint > self._last_checkpoint:
            return checkpoint_conn.create_record(self.name, self._checkpoint)
        # 重复执行就直接跳过
        logger.warn('last checkpoint {} indicates the task is already executed, bypass the execution'.format(
                self._last_checkpoint))
        return None

    def _commit(self, context, txn_id):
        """
        commit the checkpoint transaction if the execution succeeds
        :param context:
        :param txn_id:
        :return:
        """
        checkpoint_conn = context.get_connection(self.checkpoint_conn)
        checkpoint_conn.commit_record(txn_id)

    def _rollback(self, context, txn_id, err):
        """
        rollback the checkpoint transaction if execution failed
        :param context:
        :param txn_id:
        :param err:
        :return:
        """
        checkpoint_conn = context.get_connection(self.checkpoint_conn)
        checkpoint_conn.rollback_record(txn_id, err)

    def execute(self, context, **kwargs):
        """
        the execution process of the etl task
        :param context:
        :param kwargs:
        :return:
        """
        txn_id = self._start(context)
        try:
            if txn_id:
                self._result = self.execute_internal(context, **kwargs)
                self.on_commit(context, txn_id)
                self._commit(context, txn_id)
        except Exception as e:
            logger.exception(str(e))
            self._rollback(context, txn_id, e)
            self.on_rollback(context, txn_id)

    def on_commit(self, context, txn_id, **kwargs):
        pass

    def on_rollback(self, context, txn_id, **kwargs):
        pass

    def execute_internal(self, context, **kwargs):
        """
        the execution process of the task
        :param context: the executor context
        :param kwargs: the task arguments
        :return: the task result
        """
        raise NotImplementedError


class ETLTask(Task):
    @property
    def target_conn(self):
        """
        the target connection to write the result
        :return:
        """
        raise NotImplemented("The target is required")

    @property
    def target_table(self):
        """
        the target table to store the result
        :return:
        """
        return self.name

    @property
    def target_mode(self):
        """
        what to do if the target table exists, replace / append / fail
        :return:
        """
        return 'replace'

    @property
    def target_typehints(self):
        """
        a dict of column_name => datatype, to customize the data type before write target
        :return:
        """
        return {}

    @property
    def target_index(self):
        return None

    def execute_internal(self, context, **kwargs):
        """
        the internal execution process to be implemented
        :param context:
        :param kwargs:
        :return:
        """
        raise NotImplemented

    def on_commit(self, context, txn_id, **kwargs):
        target_df = self._result
        if self.target_index:
            target_df.set_index(self.target_index)
        target_conn = context.get_connection(self.target_conn)
        target_conn.store(target_df, self.target_table,
                          if_exists=self.target_mode,
                          chunksize=kwargs.get('chunksize', 10000),
                          typehints=self.target_typehints)


class SqlETLTask(ETLTask):
    @property
    def source_conn(self):
        """
        the source connection to write the result
        :return:
        """
        raise NotImplemented("The source is required")

    @property
    def etl_sql(self):
        """
        the single sql statement to process etl
        :return:
        """
        raise NotImplemented("The etl-sql is required")

    def execute_internal(self, context, **kwargs):
        source_conn = context.get_connection(self.source_conn)
        assert isinstance(source_conn, Connection)
        df = source_conn.load_query(self.etl_sql)
        return df


class APITask(Task):
    ATTR_TOTAL_ELEMENTS = 'totalElements'
    ATTR_VIEW_LABELS = 'labels'
    ATTR_EXPORT_LABELS = 'export_labels'

    def execute_internal(self, context, **kwargs):
        raise NotImplemented

    def execute(self, context, **kwargs):
        raw = self.execute_internal(context, **kwargs)
        self._attributes[APITask.ATTR_VIEW_LABELS] = self.labels
        self._attributes[APITask.ATTR_EXPORT_LABELS] = self.export_labels
        return raw

    @property
    def labels(self):
        return {}

    @property
    def export_labels(self):
        return {}



import logging

_formatter = logging.Formatter(
        '%(asctime)s program_name %(levelname)s [%(process)d] [%(filename)s:%(lineno)s]: %(message)s')

_handler = None

_console_handler = logging.StreamHandler()
_console_handler.setFormatter(_formatter)

logger = logging.getLogger()
logger.addHandler(_console_handler)
logger.setLevel(logging.DEBUG)


def setup_logging(log_path):
    global _handler
    global logger
    if _handler:
        logger.removeHandler(_handler)

    logger.debug("Logging path set to {}".format(log_path))
    _handler = logging.handlers.TimedRotatingFileHandler(log_path)
    _handler.setFormatter(_formatter)

    logger = logging.getLogger()
    logger.addHandler(_handler)
    logger.setLevel(logging.DEBUG)




from ..utils.log import logger


class Result(object):
    def __init__(self):
        self.data = None
        self.attributes = dict()

    @staticmethod
    def builder():
        return ResultBuilder()

    def content(self, labels=None):
        import pandas as pd
        if self.data is not None:
            if isinstance(self.data, pd.DataFrame):
                out_data = self.data

                nser_columns = list(out_data.select_dtypes(include=['datetime64']).keys())

                if len(nser_columns) > 0:
                    logger.warn(
                        "Detect columns {} with type that cannot be JSON serializable, which is automatically converted to string!".format(
                            list(nser_columns)))

                    for nser_column in nser_columns:
                        out_data[nser_column] = out_data[nser_column].apply(lambda x: str(x))

                if labels is not None and len(labels) > 0:
                    out_data = self.data.rename(columns=labels)
                return out_data.to_dict(orient='records')
        return self.data

    @property
    def attrs(self):
        return self.attributes

    def export(self, export_type, export_file='export', for_web=False, labels=None):
        """
        export to excel or csv
        :param export_type:
        :param export_file:
        :param for_web:
        :param labels:
        :return: the io content of the exported file
        """
        import pandas as pd
        if isinstance(self.data, pd.DataFrame):
            out_data = self.data
            if labels is not None and len(labels) > 0:
                out_data = self.data.rename(columns=labels)

            # Safe import for either Python 2.x or 3.x
            try:
                from io import BytesIO
            except ImportError:
                import StringIO as BytesIO
            io = BytesIO()

            if export_type == 'excel':
                writer = pd.ExcelWriter(io, engine='xlsxwriter') if for_web else pd.ExcelWriter(export_file,
                                                                                                engine='xlsxwriter')
                out_data.to_excel(writer, index=False)
                writer.save()
            elif export_type == 'csv':
                out_data.to_csv(io, index=False)
            else:
                raise NotImplementedError("export type {} is not supported".format(export_type))
            return io
        raise RuntimeError('the result cannot be exported')


class ResultBuilder(object):
    def __init__(self):
        self.result = Result()

    def set_data(self, data):
        self.result.data = data
        return self

    def set_attr(self, key, val):
        self.result.attributes[key] = val
        return self

    def set_attrs(self, attrs):
        if attrs is not None:
            self.result.attributes.update(attrs)
        return self

    def build(self):
        return self.result



import threading
from concurrent.futures import ThreadPoolExecutor

from tornado import gen, queues, ioloop

from ..utils.log import logger


class Engine(object):
    """
    An engine to execute etl tasks at runtime. The engine is initialized by an ETL executor context.
    """

    def __init__(self, context, detach=False):
        """
        Initialize the Engine with the executor context
        :param context: the parade context to boot engine
        :param detach: the flag to indicate the engine is executed in detached mode or not
        :return: the initialized engine
        """
        self.context = context

        # the thread pool to convert block execution of task into async process
        self.thread_pool = ThreadPoolExecutor(4)
        self.wait_queue = queues.Queue()
        self.exec_queue = queues.Queue()

        def engine_loop():
            _ioloop = ioloop.IOLoop.current()
            _ioloop.add_callback(self.daemon_loop)
            _ioloop.start()

        # if the detached mode is enabled
        # use a seperated thread to boot the io-loop
        self.loop_thread = None
        if detach:
            logger.debug("engine running in detach mode")
            self.loop_thread = threading.Thread(target=engine_loop)
            self.loop_thread.start()

    # Global lock for creating global Engine instance
    _instance_lock = threading.Lock()

    @staticmethod
    def instance(context=None, detach=False):
        """
        the static method to get or init the singleton parade-engine
        :param context: the context to init the engine
        :param detach: the flag to indicate the engine is executed in detached mode or not
        :return:
        """
        if not hasattr(Engine, "_instance"):
            with Engine._instance_lock:
                if not hasattr(Engine, "_instance"):
                    # New instance after double check
                    assert context is not None, "no context specified"
                    Engine._instance = Engine(context, detach)
        return Engine._instance

    def execute(self, task_name, **kwargs):
        """
        execute a single task with provided arguments
        :param task_name: the task name
        :param kwargs: the arguments
        :return:
        """
        assert task_name in self.context.task_dict, 'task {} not found'.format(task_name)
        task = self.context.task_dict[task_name]
        result = task.execute(self.context, **kwargs)
        return result, task.attributes

    def execute_dag(self, *task_names):
        """
        execute a set of tasks with DAG-topology into consideration
        :param task_names: the tasks to form DAG
        :return:
        """
        self.submit_dag(*task_names)
        io_loop = ioloop.IOLoop.current()
        io_loop.run_sync(self.execute_dag_ioloop)

    def submit_dag(self, *task_names):
        for task_name in task_names:
            assert task_name in self.context.task_dict, 'task {} not found'.format(task_name)

        # add to wait queue, waiting to execute
        self.wait_queue.put(set(task_names))

    @gen.coroutine
    def daemon_loop(self):
        yield self.execute_dag_ioloop()

    @gen.coroutine
    def execute_dag_ioloop(self):
        """
        the async process to execute task DAG
        :return:
        """
        executing, done = set(), set()

        @gen.coroutine
        def _produce_tasks():
            """
            the inner async procedure of task producer
            :return:
            """
            # wait until all tasks in executing queue are done
            yield self.exec_queue.join()

            # reset the *executing* and *done* set
            executing.clear()
            done.clear()

            # retrieve the task-DAG from wait-queue to exec-queue
            sched_task_names = yield self.wait_queue.get()
            for sched_task_name in sched_task_names:
                self.exec_queue.put(sched_task_name)
            self.wait_queue.task_done()

        @gen.coroutine
        def _consume_task():
            """
            the inner async procedure of task consumers
            :return:
            """
            next_task_name = yield self.exec_queue.get()
            logger.debug("pick up task [{}] ...".format(next_task_name))
            try:
                if next_task_name in executing:
                    logger.debug("task [{}] is executing, pass ...".format(next_task_name))
                    return

                next_task = self.context.task_dict[next_task_name]
                task_deps = next_task.deps
                # if len(task_deps) > 0:
                #     logger.debug(
                #         "task [{}] has {} dependant task(s), {}".format(next_task_name, len(task_deps), task_deps))
                done_deps = set(filter(lambda x: x in done, task_deps))

                if len(task_deps) == len(done_deps):
                    # all dependencies are done
                    # submit the task to threading pool to execute
                    if len(task_deps) > 0:
                        logger.debug("all dependant task(s) of task {} is done".format(next_task_name))
                    executing.add(next_task_name)

                    logger.debug("task {} start executing ...".format(next_task_name))
                    yield self.thread_pool.submit(next_task.execute, self.context)
                    logger.debug("task {} Executed successfully".format(next_task_name))
                    done.add(next_task_name)

                else:
                    # otherwise, re-put the task into the end of the queue
                    # sleep for 1 second
                    self.exec_queue.put(next_task_name)
                    yield gen.sleep(1)
            except Exception as e:
                logger.exception(str(e))
            finally:
                self.exec_queue.task_done()

        @gen.coroutine
        def consumer():
            while True:
                yield _consume_task()

        @gen.coroutine
        def producer():
            while True:
                yield _produce_tasks()

        for _ in range(2):
            consumer()

        # we use a single producer within the main-thread
        producer()

        yield self.exec_queue.join()
        assert executing == done



from flask import current_app, make_response
from flask_restful import Resource, Api

from .result import Result
from ..core.engine import Engine
from ..core.task import APITask
from . import api
from ..utils.log import logger

api = Api(api, catch_all_404s=True)


class TaskAPI(Resource):
    """
    The api blue
    print to execute etl task
    """

    def __init__(self):
        app_name = current_app.name
        app_context = current_app.config[app_name]

        self.engine = Engine(app_context)

    def post(self, task_key):
        # data in string format and you have to parse into dictionary
        from flask import request

        try:
            task_args = request.get_json()
            if not task_args:
                task_args = {}
        except Exception as e:
            logger.exception(str(e))
            task_args = {}

        task_args_ = dict(map(lambda x: (x[0], x[1][0]), filter(lambda x: len(x[1]) == 1, dict(request.args).items())))

        if task_args_:
            task_args.update(task_args_)

        self._process(task_key, **task_args)

    def get(self, task_key):
        # data in string format and you have to parse into dictionary
        from flask import request
        task_args = dict(map(lambda x: (x[0], x[1][0]), filter(lambda x: len(x[1]) == 1, dict(request.args).items())))

        self._process(task_key, **task_args)

    def _process(self, task_key, **kwargs):
        raw_result, attrs = self.engine.execute(task_key, **kwargs)
        result = Result.builder().set_data(raw_result).set_attrs(attrs).build()

        export_type = kwargs.get("_export", None)
        export_name = kwargs.get("_filename", None)
        compact = kwargs.get("_compact", None)

        if export_type is not None:
            export_labels = result.attrs.get(APITask.ATTR_EXPORT_LABELS, {})
            export_io = result.export(export_type, export_name, for_web=True, labels=export_labels)
            file_contents = export_io.getvalue()
            response = make_response(file_contents)
            response.headers["Content-Disposition"] = "attachment; filename=" + str(export_name)
            return response

        view_labels = result.attrs.get(APITask.ATTR_VIEW_LABELS, {})
        if compact is not None:
            return result.content(labels=view_labels)

        resp = {"code": 200, "content": result.content(labels=view_labels), "message": "OK"}
        resp.update(result.attrs)
        return resp


api.add_resource(TaskAPI, '/api/task/<task_key>')



#!/usr/bin/env python
# -*- coding:utf-8 -*-

from setuptools import setup, find_packages

setup(
        name='parade',
        version='0.1.1',
        description='A ETL engine for pandas-based data task',
        author='He Bai',
        author_email='bailaohe@gmail.com',
        packages=find_packages('src'),
        package_dir=({'parade': 'src/parade'}),
        zip_safe=False,
        include_package_data=True,
        package_data={'': ['*.json', '*.xml', '*.yml', '*.tpl']},
        install_requires=['pandas', 'sqlalchemy', 'requests', 'pyyaml', 'flask',
                          'flask_cors', 'flask_restful', 'psycopg2', 'mysqlclient', 'xlsxwriter',
                          'tornado'],
        entry_points={
            'console_scripts': [
                'parade = parade.cmdline:execute',
            ],
        },
        url='https://github.com/bailaohe/parade',
        download_url='https://github.com/bailaohe/parade/tarball/0.1',
)



"""Helper functions to handle object search in module"""
import inspect
from importlib import import_module
from pkgutil import iter_modules

import re


def load_object(path):
    """Load an object given its absolute object path, and return it.

    object can be a class, function, variable or an instance.
    path ie: 'scrapy.downloadermiddlewares.redirect.RedirectMiddleware'
    """

    try:
        dot = path.rindex('.')
    except ValueError:
        raise ValueError("Error loading object '%s': not a full path" % path)

    module, name = path[:dot], path[dot + 1:]
    mod = import_module(module)

    try:
        obj = getattr(mod, name)
    except AttributeError:
        raise NameError("Module '%s' doesn't define any object named '%s'" % (module, name))

    return obj


def walk_modules(path):
    """Loads a module and all its submodules from the given module path and
    returns them. If *any* module throws an exception while importing, that
    exception is thrown back.

    For example: walk_modules('scrapy.utils')
    """

    mods = []
    mod = import_module(path)
    mods.append(mod)
    if hasattr(mod, '__path__'):
        for _, subpath, ispkg in iter_modules(mod.__path__):
            fullpath = path + '.' + subpath
            if ispkg:
                mods += walk_modules(fullpath)
            else:
                submod = import_module(fullpath)
                mods.append(submod)
    return mods


def iter_classes(base_class, *modules, class_filter=None):
    """
    iterate the class within specified module paths satisfying the class filter
    :param base_class: the base class of the one to search
    :param modules: the module paths
    :param class_filter: the class filter
    :return:
    """
    for root_module in modules:
        for module in walk_modules(root_module):
            for obj in vars(module).values():
                if inspect.isclass(obj) and issubclass(obj, base_class) and obj.__module__ == module.__name__:
                    if not class_filter or class_filter(obj):
                        yield obj


def get_class(name, base_class, *modules):
    """
    get the first class with name as specified (with name-property or module name) in module paths
    :param name: the name to search for the class
    :param base_class: the base class of the one to search
    :param modules: the module paths
    :return:
    """
    for cls in iter_classes(base_class, *modules, class_filter=lambda x: x.__module__.split('.')[-1] == name):
        return cls
    return None



class ConfigObject(object):
    def __init__(self, value):
        if not isinstance(value, dict):
            raise TypeError('Can only accept dict')
        self._dict = value

    def get(self, path=None):

        if not path:
            return self._dict

        tokens = path.split(".")

        conf_node = self._dict
        for item in tokens:
            if item not in conf_node:
                raise RuntimeError('load config [{}] failed'.format(path))
            conf_node = conf_node[item]

        if isinstance(conf_node, dict):
            return ConfigObject(conf_node)
        return str(conf_node)

    def to_dict(self):
        return self._dict

    def __getitem__(self, path):
        return self.get(path)

    def __str__(self):
        return self._dict


class ConfigRepo(object):
    def __init__(self, url, **kwargs):
        self.url = url
        self.kwargs = kwargs

    def load(self, app_name, profile='default', **kwargs):
        return ConfigObject(self.load_internal(app_name, profile, **kwargs))

    def load_internal(self, app_name, profile='default', **kwargs):
        raise NotImplementedError



class Datasource(object):
    """
    The data source object. The object does not maintain any stateful information.
    """

    def __init__(self, driver, **kwargs):
        self.driver = driver
        self.attributes = kwargs

    @property
    def protocol(self):
        return self.attributes['protocol'] if 'protocol' in self.attributes else None

    @property
    def host(self):
        return self.attributes['host'] if 'host' in self.attributes else None

    @property
    def port(self):
        return self.attributes['port'] if 'port' in self.attributes else None

    @property
    def user(self):
        return self.attributes['user'] if 'user' in self.attributes else None

    @property
    def password(self):
        return self.attributes['password'] if 'password' in self.attributes else None

    @property
    def db(self):
        return self.attributes['db'] if 'db' in self.attributes else None

    @property
    def uri(self):
        return self.attributes['uri'] if 'uri' in self.attributes else None


class Connection(object):
    """
    The connection object, which is opened with data source and its implementation is also
    related to the context
    """

    def __init__(self, datasource):
        self.datasource = datasource

    def load(self, table, **kwargs):
        raise NotImplementedError

    def load_query(self, query, **kwargs):
        raise NotImplementedError

    def store(self, df, table, **kwargs):
        raise NotImplementedError

    @property
    def accept(self):
        raise NotImplementedError

    """
        -- mysql dialect
    CREATE TABLE _task_record (
        `id` bigint(20) unsigned NOT NULL AUTO_INCREMENT,
        `task` varchar(64) NOT NULL,
        `checkpoint` timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP,
        `create_time` timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP,
        `commit_time` timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP,
        `update_time` timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP ON UPDATE CURRENT_TIMESTAMP,
        `status` int(11) NOT NULL DEFAULT 0, -- 0: running, 1: committed, 2: failed
        `message` varchar(256) NOT NULL DEFAULT 'OK',
         PRIMARY KEY (`id`),
         KEY `idx_task_create` (`task`,`create_time`)
    );

    -- postgresql dialect
    CREATE TABLE public._task_record (
        id SERIAL PRIMARY KEY NOT NULL,
        task CHARACTER VARYING(64) NOT NULL,
        checkpoint TIMESTAMP WITHOUT TIME ZONE NOT NULL DEFAULT now(),
        create_time TIMESTAMP WITHOUT TIME ZONE NOT NULL DEFAULT now(),
        commit_time TIMESTAMP WITHOUT TIME ZONE NOT NULL DEFAULT now(),
        update_time TIMESTAMP WITHOUT TIME ZONE NOT NULL DEFAULT now(),
        status INTEGER NOT NULL DEFAULT 0,
        message CHARACTER VARYING NOT NULL DEFAULT 'OK'::character varying
    );
    CREATE INDEX idx_task_create ON _task_record USING BTREE (task, create_time);
    """

    @property
    def can_produce(self):
        raise NotImplementedError

    def last_record(self, task_name):
        raise NotImplementedError

    def create_record(self, task_name, new_checkpoint):
        raise NotImplementedError

    def commit_record(self, txn_id):
        raise NotImplementedError

    def rollback_record(self, txn_id, err):
        raise NotImplementedError



from . import Scheduler
from ..core.engine import Engine
from ..utils.log import logger


class ParadeScheduler(Scheduler):
    """
    The scheduler to support task-flow scheduling
    """

    def schedule(self, *tasks, cron=None, flow_name=None):
        """
        schedule the task-flow
        :param flow_name: the flow name
        :param cron: the cron string to schedule the flow
        :return:
        """
        if cron:
            raise NotImplementedError('default scheduler cannot support cron')
        engine = Engine(self.context)

        logger.debug('prepare to schedule task(s) {}'.format(tasks))

        if tasks and len(tasks) > 0:
            engine.execute_dag(*tasks)

        return None

    def unsched(self, flow_name):
        """
        unschedule the task-flow
        :param flow_name: the flow name
        :return:
        """
        pass

